﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using HeuristicLab.DataImporter.Data.CommandBase;
using HeuristicLab.DataImporter.Data.Model;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
namespace HeuristicLab.DataImporter.Data.Command {
  [StorableClass]
  public class SortCommand : ColumnGroupCommandBase {
    [Storable]
    private bool addToSortedColumns;
    [Storable]
    private int sortColumnIndex;

    private ICollection<int> oldSortedColumnIndices;
    private IEnumerable<SortOrder> oldSortOrdersForColumns;
    private int[] indices;

    private SortCommand()
      : base(null, string.Empty) {
    }

    public SortCommand(DataSet dataSet, string columnGroupName, int sortColumnIndex) :
      this(dataSet, columnGroupName, sortColumnIndex, false) {
    }

    public SortCommand(DataSet dataSet, string columnGroupName, int sortColumnIndex, bool addToSortedColumns)
      : base(dataSet, columnGroupName) {
      this.addToSortedColumns = addToSortedColumns;
      this.sortColumnIndex = sortColumnIndex;
    }

    public override void Execute() {
      base.Execute();
      oldSortedColumnIndices = new List<int>(ColumnGroup.SortedColumnIndexes);
      oldSortOrdersForColumns = ColumnGroup.SortOrdersForColumns.ToList();

      if (!addToSortedColumns) {
        ColumnGroup.SortedColumnIndexes.Clear();
        for (int i = 0; i < ColumnGroup.Columns.Count(); i++)
          if (i != sortColumnIndex)
            ColumnGroup.Columns.ElementAt(i).SortOrder = SortOrder.None;
      }
      if (!ColumnGroup.SortedColumnIndexes.Contains(sortColumnIndex))
        ColumnGroup.SortedColumnIndexes.Add(sortColumnIndex);
      ICollection<int> tempSortedColumnIndices = new List<int>(ColumnGroup.SortedColumnIndexes);

      RowComparer rowComparer = new RowComparer(ColumnGroup, sortColumnIndex);
      indices = Enumerable.Range(0, ColumnGroup.RowCount).ToArray();
      Array.Sort(indices, rowComparer);

      List<ColumnBase> newColumns = Sort();
      ColumnGroup.ClearColumns();
      ColumnGroup.AddColumns(newColumns);
      ColumnGroup.SortedColumnIndexes = tempSortedColumnIndices;
      ColumnGroup.FireChanged();
      this.ColumnGroup = null;
    }

    public override void UndoExecute() {
      base.UndoExecute();
      indices = Enumerable.Range(0, ColumnGroup.RowCount).OrderBy<int, int>(i => indices[i]).ToArray();
      List<ColumnBase> newColumns = Sort();
      ColumnGroup.ClearColumns();
      ColumnGroup.AddColumns(newColumns);
      ColumnGroup.SortedColumnIndexes = this.oldSortedColumnIndices;
      ColumnGroup.SortOrdersForColumns = oldSortOrdersForColumns;
      indices = null;
      oldSortedColumnIndices = null;
      oldSortOrdersForColumns = null;
      ColumnGroup.FireChanged();
      this.ColumnGroup = null;
    }

    public override string Description {
      get { return "Sort ColumnGroup"; }
    }

    private List<ColumnBase> Sort() {
      List<ColumnBase> newColumns = new List<ColumnBase>();
      ColumnBase newColumn;
      ColumnBase oldColumn;

      for (int col = 0; col < ColumnGroup.Columns.Count(); col++) {
        oldColumn = ColumnGroup.GetColumn(col);
        newColumn = oldColumn.CreateCopyOfColumnWithoutValues();
        newColumn.SortOrder = oldColumn.SortOrder;
        if (col == sortColumnIndex) {
          if (newColumn.SortOrder == SortOrder.None || newColumn.SortOrder == SortOrder.Descending)
            newColumn.SortOrder = SortOrder.Ascending;
          else
            newColumn.SortOrder = SortOrder.Descending;
        }
        newColumns.Add(newColumn);
      }

      for (int j = 0; j < indices.Length; j++) {
        for (int col = 0; col < ColumnGroup.Columns.Count(); col++)
          newColumns[col].AddValue(ColumnGroup.Columns.ElementAt(col).GetValue(indices[j]));
      }
      return newColumns;
    }

    private class RowComparer : IComparer<int> {
      private ColumnGroup columnGroup;
      private int sortColumnIndex;
      public RowComparer(ColumnGroup columnGroup, int sortColumnIndex) {
        this.columnGroup = columnGroup;
        this.sortColumnIndex = sortColumnIndex;
      }

      #region IComparer<int> Members
      public int Compare(int i, int j) {
        IEnumerator<int> e = columnGroup.SortedColumnIndexes.GetEnumerator();
        int result = 0;
        while (result == 0 && e.MoveNext()) {
          IComparable x = columnGroup.GetColumn(e.Current).GetValue(i);
          IComparable y = columnGroup.GetColumn(e.Current).GetValue(j);
          if (x == null) {
            result = -1;
            if (y == null)
              result = 0;
          } else
            result = x.CompareTo(y);
          if (columnGroup.SortOrdersForColumns.ElementAt(e.Current) == SortOrder.Descending)
            result *= -1;
          if (e.Current == sortColumnIndex && columnGroup.Columns.ElementAt(e.Current).SortOrder != SortOrder.None)
            result *= -1;
        }
        return result;
      }
      #endregion
    }
  }
}
